
import asyncio
from datetime import datetime
from jbg_util import binance_models
from jbg_util.basic_util import now
from jbg_util import trader_models

PriceTraderData,BalanceTraderData,OrderTraderData,AssetSymbol,OrderSide,OrderStatus,OrderType = trader_models.PriceTraderData,trader_models.BalanceTraderData,trader_models.OrderTraderData,trader_models.AssetSymbol,trader_models.OrderSide,trader_models.OrderStatus,trader_models.OrderType
PriceSocketData,BalanceSocketData,OrderSocketData,SocketEventType = binance_models.PriceSocketData,binance_models.BalanceSocketData,binance_models.OrderSocketData,binance_models.SocketEventType
class BinanceMockSocketDataUtil:
    @staticmethod
    def get_mock_price_socket_data(price : float = 16812.9811) -> PriceSocketData:
        price_data = PriceTraderData(
            event_type=SocketEventType.price_update,
            event_datetime=now(),
            symbol=AssetSymbol.BTC,
            price=price
        )
        return price_data.to_socket_data()

    @staticmethod
    def get_mock_balance_socket_data(btc_free : float=.00397067,btc_locked: float=.90397067,usd_free: float=30000.5623,usd_locked: float=2000.6754) -> BalanceSocketData:
        balance_data = BalanceTraderData(
            event_type=SocketEventType.balance_update,
            event_datetime=now(),
            btc_free=btc_free,
            btc_locked=btc_locked,
            usd_free=usd_free,
            usd_locked=usd_locked
        )
        return balance_data.to_socket_data()

    @staticmethod
    def get_mock_order_socket_data(order_id:int=1648660237,side:str=OrderSide.buy,qty:float=.01,filled_qty:float=.01,price:float=17200.5203,
        status:str=OrderStatus.filled,create_datetime:datetime=None,order_type:str=OrderType.limit) -> OrderSocketData:
        order_data = OrderTraderData(
            event_type=SocketEventType.order_update,
            event_datetime=now(),
            symbol=AssetSymbol.BTC,
            order_id=order_id,
            side=side,
            qty=qty,
            filled_qty=qty,
            price=price,
            status=status,
            order_type=order_type,
            create_datetime=create_datetime or now()
        )
        return order_data.to_socket_data()

    @staticmethod
    def get_static_mock_recv_fxn(delay,mock_data):
        async def mock_recv_fxn(self,socket_stream):
            await asyncio.sleep(delay)
            return mock_data
        return mock_recv_fxn

    @staticmethod
    def get_alternating_mock_recv_fxn(delays,mock_datas):
        global alternating_call_ct
        alternating_call_ct=-1
        async def mock_recv_fxn(self,socket_stream):
            global alternating_call_ct
            alternating_call_ct+=1
            idx = alternating_call_ct%len(mock_datas)
            await asyncio.sleep(delays[idx])
            return mock_datas[idx]
        return mock_recv_fxn

    @staticmethod
    def get_ending_mock_recv_fxn(delays,mock_datas):
        global ending_call_ct
        ending_call_ct=-1
        async def mock_recv_fxn(self,socket_stream):
            global ending_call_ct
            ending_call_ct+=1
            if ending_call_ct>=len(mock_datas):
                await asyncio.sleep(3600)
                return None
            await asyncio.sleep(delays[ending_call_ct])
            return mock_datas[ending_call_ct]
        return mock_recv_fxn





from datetime import datetime
from jbg_util import file_util
from jbg_util import data_converter
from jbg_util import binance_models
import os

BinanceApiKeys = binance_models.BinanceApiKeys
DataConverter = data_converter.DataConverter
FileManager = file_util.FileManager
def binance_timestamp_to_dte(ts : int):
    return datetime.utcfromtimestamp(ts/1000)

def dte_to_binance_timestamp(dte : datetime):
    return dte.timestamp()*1000

def get_binance_api_keys() -> BinanceApiKeys:
    fu = FileManager()

    env = os.getenv('JBG_ENV')
    file_path='/Users/jgardner/Documents/personal/.binance_api_keys'
    if env=='EC2':
        file_path='~/.bnb_api_keys'
    d = fu.get_obj_from_file(file_path)
    dc = DataConverter()
    keys = dc.get_instance_from_dict(d,BinanceApiKeys)
    return keys



from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar

from jbg_util import data_converter

DataConverter = data_converter.DataConverter

class SocketEventType:
    order_update = 'executionReport'
    balance_update = 'outboundAccountPosition'
    price_update = '24hrTicker'

class SocketChannel:
    # dflt = 0
    price = 'PRICE'
    balance = 'BALANCE'
    order = 'ORDER'
@dataclass
class BinanceApiKeys:
    api_key : str
    api_secret : str

@dataclass
class SocketData:
    e : str #Event type
    E : int # event time

    data_converter : ClassVar[DataConverter]

    @staticmethod
    def get_data_converter():
        return DataConverter()

    @staticmethod
    def from_raw_socket_data(raw_data):
        return SocketData.data_converter.get_instance_from_dict(raw_data)
@dataclass
class PriceSocketData(SocketData):
    s : str # symbol
    c : float # last price

    @staticmethod
    def from_raw_socket_data(raw_data):
        return PriceSocketData.data_converter.get_instance_from_dict(raw_data,PriceSocketData)
# {
#   "e": "24hrTicker",  // Event type
#   "E": 123456789,     // Event time
#   "s": "BNBBTC",      // Symbol
#   "p": "0.0015",      // Price change
#   "P": "250.00",      // Price change percent
#   "w": "0.0018",      // Weighted average price
#   "x": "0.0009",      // First trade(F)-1 price (first trade before the 24hr rolling window)
#   "c": "0.0025",      // Last price
#   "Q": "10",          // Last quantity
#   "b": "0.0024",      // Best bid price
#   "B": "10",          // Best bid quantity
#   "a": "0.0026",      // Best ask price
#   "A": "100",         // Best ask quantity
#   "o": "0.0010",      // Open price
#   "h": "0.0025",      // High price
#   "l": "0.0010",      // Low price
#   "v": "10000",       // Total traded base asset volume
#   "q": "18",          // Total traded quote asset volume
#   "O": 0,             // Statistics open time
#   "C": 86400000,      // Statistics close time
#   "F": 0,             // First trade ID
#   "L": 18150,         // Last trade Id
#   "n": 18151          // Total number of trades
# }

@dataclass
class AssetBalance:
    a : str # Asset
    f : float #free
    l : float #locked
@dataclass
class BalanceSocketData(SocketData):
    B : list[AssetBalance] #balances

    data_converter : ClassVar[DataConverter]
    @staticmethod
    def get_data_converter():
        dc =DataConverter()

        def convert_balances(raw_balance_data):
            return dc.get_instances_from_arr_of_dicts(raw_balance_data,AssetBalance)
        dc.conversion_fxn_map['list[AssetBalance]']=convert_balances
        return dc

    @staticmethod
    def from_raw_socket_data(raw_data):
        return BalanceSocketData.data_converter.get_instance_from_dict(raw_data,BalanceSocketData)
# {
#   "e": "outboundAccountPosition", //Event type
#   "E": 1564034571105,             //Event Time
#   "u": 1564034571073,             //Time of last account update
#   "B": [                          //Balances Array
#     {
#       "a": "ETH",                 //Asset
#       "f": "10000.000000",        //Free
#       "l": "0.000000"             //Locked
#     }
#   ]
# }

@dataclass
class OrderSocketData(SocketData):
    s : str #Symbol
    i : int #orderID
    S : str #side
    o : str #orderType
    q : float #orderQty
    p : float #price
    z : float #cumulative filled qty
    X : str #orderStatus
    O : int #creationTime

    @staticmethod
    def from_raw_socket_data(raw_data):
        return OrderSocketData.data_converter.get_instance_from_dict(raw_data,OrderSocketData)
# {
#   "e": "executionReport",        // Event type
#   "E": 1499405658658,            // Event time
#   "s": "ETHBTC",                 // Symbol
#   "c": "mUvoqJxFIILMdfAW5iGSOW", // Client order ID
#   "S": "BUY",                    // Side
#   "o": "LIMIT",                  // Order type
#   "f": "GTC",                    // Time in force
#   "q": "1.00000000",             // Order quantity
#   "p": "0.10264410",             // Order price
#   "P": "0.00000000",             // Stop price
#   "d": 4,                        // Trailing Delta; This is only visible if the order was a trailing stop order.
#   "F": "0.00000000",             // Iceberg quantity
#   "g": -1,                       // OrderListId
#   "C": "",                       // Original client order ID; This is the ID of the order being canceled
#   "x": "NEW",                    // Current execution type
#   "X": "NEW",                    // Current order status
#   "r": "NONE",                   // Order reject reason; will be an error code.
#   "i": 4293153,                  // Order ID
#   "l": "0.00000000",             // Last executed quantity
#   "z": "0.00000000",             // Cumulative filled quantity
#   "L": "0.00000000",             // Last executed price
#   "n": "0",                      // Commission amount
#   "N": null,                     // Commission asset
#   "T": 1499405658657,            // Transaction time
#   "t": -1,                       // Trade ID
#   "I": 8641984,                  // Ignore
#   "w": true,                     // Is the order on the book?
#   "m": false,                    // Is this trade the maker side?
#   "M": false,                    // Ignore
#   "O": 1499405658657,            // Order creation time
#   "Z": "0.00000000",             // Cumulative quote asset transacted quantity
#   "Y": "0.00000000",             // Last quote asset transacted quantity (i.e. lastPrice * lastQty)
#   "Q": "0.00000000",             // Quote Order Qty
#   "D": 1668680518494,            // Trailing Time; This is only visible if the trailing stop order has been activated.
#   "j": 1,                        // Strategy ID; This is only visible if the strategyId parameter was provided upon order placement
#   "J": 1000000                   // Strategy Type; This is only visible if the strategyType parameter was provided upon order placement
#   "W": 1499405658657,            // Working Time; This is only visible if the order has been placed on the book.
#   "V": "NONE"                    // SelfTradePreventionMode
# }


for cls in [SocketData,BalanceSocketData,OrderSocketData,PriceSocketData]:
    cls.data_converter = cls.get_data_converter()


from types import FunctionType
import asyncio
class DestinationType:
    local_fxn = 1

class Subscriber:
    def __init__(self,destination,destination_type : DestinationType):
        self.destination = destination
        self.destination_type = destination_type

class PubSubMediator:
    def __init__(self):
        self.subscribers_by_channel = {}

    async def publish(self, data, channel : str):
        if channel not in self.subscribers_by_channel:
            return
        local_fxns_to_call=[]
        for subscriber in self.subscribers_by_channel[channel]:
            if subscriber.destination_type == DestinationType.local_fxn:
                local_fxns_to_call.append(subscriber.destination)
        await asyncio.gather(*[fxn(data,channel) for fxn in local_fxns_to_call])

    def subscribe_local_fxn(self, destination : FunctionType, channel : str):
        subscriber = Subscriber(destination,DestinationType.local_fxn)
        if channel not in self.subscribers_by_channel:
            self.subscribers_by_channel[channel]=[]
        self.subscribers_by_channel[channel].append(subscriber)




from __future__ import annotations
import asyncio
from binance import AsyncClient, BinanceSocketManager
from jbg_util import binance_models
from jbg_util import pub_sub_mediator
from jbg_util import logging_util
from jbg_util import data_converter
from json import dumps

Logger = logging_util.Logger
DataConverter = data_converter.DataConverter
PubSubMediator = pub_sub_mediator.PubSubMediator
SocketData,BalanceSocketData,PriceSocketData,OrderSocketData,SocketChannel,SocketEventType,BinanceApiKeys = \
    binance_models.SocketData,binance_models.BalanceSocketData,binance_models.PriceSocketData,binance_models.OrderSocketData,binance_models.SocketChannel,binance_models.SocketEventType,binance_models.BinanceApiKeys

# TODO: don't make socket connect in tests with mockData!
class SocketManager:
    def __init__(self,logger : Logger = None):
        self.logger = logger or Logger(file_path='./logs/socket_manager.log',should_print_to_screen=True).empty_log()
        self.socket_streams : list[BinanceSocketStream]=[]

    async def start_sockets(self,api_keys : BinanceApiKeys, pub_sub_mediator : PubSubMediator,socket_classes):
        tasks = []
        for cls in socket_classes:
            socket_stream = cls(pub_sub_mediator,self.logger)
            task = asyncio.create_task(self.start_socket(api_keys,socket_stream))

            self.socket_streams.append(socket_stream)
            tasks.append(task)
        rslts = await asyncio.gather(*tasks)

    async def start_socket(self,api_keys : BinanceApiKeys,socket_stream : BinanceSocketStream):
        await socket_stream.init(api_keys)
        await socket_stream.open_socket()

    def stop_sockets(self):
        for socket_stream in self.socket_streams:
            socket_stream.close_socket()

class BinanceSocketStream:
    data_converter = DataConverter()

    def __init__(self, pub_sub_mediator : PubSubMediator, logger : Logger, name : str=None):
        self.pub_sub_mediator=pub_sub_mediator
        self.name=name
        self.close_socket_event = asyncio.Event()
        self.logger = logger

    async def open_socket(self):
        raise NotImplementedError("A generic socket stream cannot be opened.")

    async def init(self, api_keys : BinanceApiKeys):
        self.client = await AsyncClient.create(api_key=api_keys.api_key, api_secret=api_keys.api_secret,tld='us')
        self.binance_socket_manager = BinanceSocketManager(self.client)

    async def recv(self,socket_stream):
        return await socket_stream.recv()

    async def listen(self):
        self.logger.info(f'Opened the {self.name} socket connection...')
        try:
            async with self.ts as socket_stream:
                while True:
                    completed,pending = await asyncio.wait([self.recv(socket_stream),self.wait_until_close_socket_event()],return_when=asyncio.FIRST_COMPLETED)
                    if self.close_socket_event.is_set():
                        self.logger.info(f'Close socket event has occurred on socket_stream={self.name}')
                        return
                    for task in completed:
                        # print('\n\nTASK\n',task)
                        # print('\n\nRESULT\n',task.result())
                        await self.publish(task.result())
        except Exception as e:
            raise e
        finally:
            self.logger.info(f'Closing the {self.name} connection...')
            await self.client.close_connection()

    async def wait_until_close_socket_event(self):
        await self.close_socket_event.wait()

    def close_socket(self):
        self.close_socket_event.set()

    async def publish(self,data):
        self.logger.info(dumps(data))
        data,channel=self.parse_data(data)
        if self.pub_sub_mediator is None:
            return
        await self.pub_sub_mediator.publish(data,channel)

    @staticmethod
    def parse_data(data) -> SocketData:
        return data,None

class BinanceUserSocketStream(BinanceSocketStream):
    def __init__(self, pub_sub_mediator : PubSubMediator, logger : Logger):
        super().__init__(pub_sub_mediator, logger, name='user')

    async def open_socket(self):
        self.ts = self.binance_socket_manager.user_socket()
        await self.listen()

    @staticmethod
    def parse_data(data) -> SocketData:
        event_type=data['e']
        if event_type ==SocketEventType.order_update:
            data = OrderSocketData.from_raw_socket_data(data)
            return data,SocketChannel.order
        if event_type ==SocketEventType.balance_update:
            data = BalanceSocketData.from_raw_socket_data(data)
            return data,SocketChannel.balance
        raise ValueError(f'unrecognized event type = {event_type}',data)

class BinanceTickerSocketStream(BinanceSocketStream):
    def __init__(self, pub_sub_mediator : PubSubMediator, logger : Logger):
        super().__init__(pub_sub_mediator, logger, name='ticker')

    async def open_socket(self):
        self.ts = self.binance_socket_manager.symbol_ticker_socket('BTCUSD')
        await self.listen()

    @staticmethod
    def parse_data(data) -> SocketData:
        data = PriceSocketData.from_raw_socket_data(data)
        return data,SocketChannel.price